import sqlalchemy
import pyodbc
import pandas as pd

"""
Created 05.07.2021
r
:Author: Andreas Rystad
"""


class SQLConnection:
    """
    Class for connecting to a database and server

    The following guide was used:
    https://medium.com/@anushkamehra16/connecting-to-sql-database-using-sqlalchemy-in-python-2be2cf883f85
    """
    connection = None
    cursor = None
    engine = None
    server_name = None
    db_name = None

    def __init__(self, server_name, db_name):
        """
        Method initialising class with name of server and database
        :param server_name: String with server name
        :param db_name: String with DB name
        """
        self.server_name = server_name
        self.db_name = db_name
        params = "DRIVER={ODBC Driver 17 for SQL Server}; SERVER=%s; DATABASE=%s; Trusted_Connection=Yes" % (
            server_name, db_name)
        self.connection = pyodbc.connect(params)
        self.cursor = self.connection.cursor()
        self.engine = sqlalchemy.create_engine('mssql+pyodbc://%s/%s?driver=SQL+Server' % (server_name, db_name))

    def execute(self, sql_statement):
        """
        Method for executing a SQL statement
        :param: sql_statement: SQL statement to be executed
        :return: result with sqlalchemy.engine.cursor.LegacyCursorResult object. (can call .rowcount)
        """
        result = self.engine.execute(sql_statement)
        return result

    def read_to_df(self, sql_statement):
        """
        Method for reading from database and turning result to dataframe
        :param sql_statement: SQL statement to be executed
        :return: pandas.Dataframe with information from sql statement
        """
        sql_result = pd.read_sql_query(sql_statement, self.engine)
        df = pd.DataFrame(sql_result)
        return df

    def get_connection(self):
        """
        Method for getting connection
        :return: engine and connection
        """
        return self.engine, self.connection

    def get_names(self):
        """
        Method for getting server and database name
        :return:
        """
        return self.server_name, self.db_name


def make_connection(server_name, db_name):
    """
    Method for creating a new connection. Can be used if one doesnt wish to have an instance
    :param server_name: name of server
    :param db_name: name of database
    :return: connection, cursor, engine
    """
    params = "DRIVER={ODBC Driver 17 for SQL Server}; SERVER=%s; DATABASE=%s; Trusted_Connection=Yes" % (
        server_name, db_name)
    connection = pyodbc.connect(params)
    cursor = connection.cursor()
    engine = sqlalchemy.create_engine('mssql+pyodbc://%s/%s?driver=SQL+Server' % (server_name, db_name))
    return connection, cursor, engine



import pandas as pd
import resqlconnection
from datetime import datetime


# EXAMPLE 1
table_name="testtable"
schema_name="crawlertest"
server_name= "TOR"
db_name = "RECrawlers"

conn = resqlconnection.ReSqlConnection(server_name, db_name)

input_list = [["This is a test",1,1.5,datetime.now(),5], ["also a test",3,1.5,datetime(2015,1,5),10]]

col = ["Str","num","float","Time","value"] # The names of the columns as they will be in the db, a crawl date column will automatically be added

dataframe = pd.DataFrame(input_list,columns=col)

conn.push_data(table_name, schema_name, dataframe, dataframe.columns, ["value"])

# The package also allows you to get information about tables from the database
conn.get_table_info(table_name, schema_name)

conn.preview_table(table_name, schema_name)





#pushing data
#input_list2 = ["String",2,2.9,None,2]
#conn.push_data(table_name,schema_name,input_list,col,["value"],False)                       #list in list
#conn.push_data(table_name,schema_name,input_list2,col,["value"],False)                      #list
            #dataframe

#conn.get_all_tables()

#example 2 of use:
"""
list = []
for i in range(1,3):
    list.append(["Norway",pd.Timestamp.date(pd.Timestamp(year=2021,month=1,day=i)),i])
for i in range(1,3):
    list.append(["Sweden",pd.Timestamp.date(pd.Timestamp(year=2021,month=1,day=i)),i+2])

conn.push_data("test","example",list,["Country","Date","Value"],["Value"],True)
 """


#example 3, only using sql_connection class
#connection, cursor, engine = resqlconnection.sql_connection.make_connection(server_name,db_name)
# use for whatever you need, e.g
#sql_query = f"SELECT * FROM {schema_name}.{table_name}"
#engine.execute(sql_query)





from resqlconnection import methods, sql_connection

"""
Created 05.07.2021

:Author: Andreas Rystad
"""


class ReSqlConnection:
    """
    Class used for connecting and communicating with SQL

    """
    sql_conn = None
    method = None

    def __init__(self, server_name, db_name):
        """
        Method for initialising class. Makes a connection with database
        :param server_name: String with name of server
        :param db_name: String with name of database
        """
        try:
            self.sql_conn = sql_connection.SQLConnection(server_name, db_name)
            self.method = methods
            print("Connection created")
            self.method=methods.Method(255)
        except Exception as e:
            print(str(e))
            raise Exception(f"something went wrong with creating a connection with {server_name} and {db_name}")

    def get_table_info(self, table_name, schema_name):
        """
        Method for returning information about a certain table. Includes columns, data types and number of rows
        :param table_name: Name of table
        :param schema_name: Name of schema
        :return: Dataframe with columns and data types
        """
        return self.method.get_table_info(table_name, schema_name, self.sql_conn)

    def get_all_tables(self):
        """
        method for returning all tables in selected database
        :return: tables in DB
        """
        return self.method.get_all_tables_info(self.sql_conn)

    def preview_table(self, table_name, schema_name, nr_of_lines=5):
        """
        Method that prints first rows of a table
        :param table_name: Name of table
        :param schema_name: Name of schema
        :param nr_of_lines: Number of lines to be printed. Default value is 5
        :return: preview of table
        """
        return self.method.preview_table(table_name, schema_name, nr_of_lines, self.sql_conn)

    def push_data(self, table_name, schema_name, input_data, column_name_list, value_columns,
                  crawldate_column='CrawlDate', replace=False,
                  ignore_duplicates_check=False, recipients=None):
        """
        Pushes data to SQL server. Also appends a CrawlDate column if it does not exist
        :param table_name: String with name of table
        :param schema_name: String with name of schema
        :param input_data: Data to be pushed to database
        :param column_name_list: List with name of columns. Only important if input is on list format
        :param value_columns: List with name of columns where values are. These will be ignored when looking for duplicates
        :param crawldate_column: String describing name of crawl_date column. Default is 'CrawlDate'
        :param replace: Boolean describing if replacing old values or discarding new ones when a duplicate happens.
        :param ignore_duplicates_check True if you want no duplicate check to run
        :param recipients list or string with emails that will receive message if method fails
        """
        return self.method.push_data(table_name, schema_name, input_data, column_name_list, value_columns,
                                     crawldate_column, replace, ignore_duplicates_check, self.sql_conn, recipients)

    def remove_duplicates(self, table_name, schema_name, replace, value_columns):
        """
        Method for removing duplicate rows from SQL database
        :param table_name: String with name of table
        :param schema_name: String with name of schema
        :param replace: Boolean describing if replacing old values or discarding new ones when a duplicate happens
        :param value_columns: Columns that contains value. Will be ignored when looking for duplicates
        :param crawldate_column: String describing name of crawl_date column. Default is 'CrawlDate'
        :return:
        """
        return self.method.remove_duplicates(table_name, schema_name, replace, value_columns, self.sql_conn)

    def send_email(self, recipients, message,subject="Crawler Failed"):
        """
        :param recipients email addresses, string or list
        :param message to be written in email. The subject is always "Crawler failed"
        """
        return self.method.send_email(self.sql_conn, recipients, message,subject)

    def set_max_varchar(self, max_varchar_size):
        """
        :param max_varchar_size size of maximum length of varchar
        """
        self.method.set_max_varchar(max_varchar_size)

    def get_newest_date(self, table_name, schema_name, date_column_name):
        """ 
        Getting the newest date present
        :param date_column_name: name of column in database that holds date
        """
        return self.method.get_newest_date(table_name, schema_name, date_column_name, self.sql_conn)

    def get_row_count(self, table_name, schema_name):
        """ 
        Getting the number of rows
        """
        return self.method.get_row_count(table_name, schema_name, self.sql_conn)



from setuptools import setup





def readme():
    with open('README.rst') as file:
        return file.read()


setup(name='resqlconnection',
      version='0.12',
      description='The RE crawling tool',
      author="Andreas Rystad",
      author_email="TechnologySupport@rystadenergy.com",
      license="MIT",  # ?
      packages=["RESqlConnection"],
      install_requires=[
          'pandas>1.3.0', 'numpy>1.21.0', 'SQLAlchemy>1.4.20', 'pyodbc>4.0.30', 'pymssql>2.2.1', 'setuptools>56.0.0'
      ],
      zip_safe=False)



import math

import numpy as np
import pandas
import pandas as pd
import sqlalchemy as sal

"""
Created 05.07.2021

:Author: Andreas Rystad
"""


class Method:
    max_varchar = 255

    def __init__(self, varchar_size):
        """
        :param max_varchar_size size of maximum length of varchar
        """
        self.set_max_varchar(varchar_size)


    def get_table_columns(self, table_name, schema_name, sql_conn):
        """
        method for returning table columns
        :parameter table_name name of table
        :parameter schema_name name of schema
        :parameter sql_conn instance of SQLConnection class

        :returns Dataframe with columns in SQL table
        """
        columns = sql_conn.read_to_df("SELECT COLUMN_NAME,DATA_TYPE FROM INFORMATION_SCHEMA.COLUMNS " +
                                      f"WHERE TABLE_NAME = '{table_name}' AND TABLE_SCHEMA ='{schema_name}'")
        return columns


    def get_table_info(self, table_name, schema_name, sql_conn):
        """
        Prints table information for a given table name with columns, datatypes and nr of rows
        :parameter table_name name of table
        :parameter schema_name name of schema
        :parameter sql_conn instance of SQLConnection class

        :returns Dataframe with columns and types in SQL table
        """
        columns = self.get_table_columns(table_name, schema_name, sql_conn)
        print(self.format_string(table_name + "\n" + str(columns)))
        number_of_rows = sql_conn.read_to_df(f"Select Count(*) FROM {schema_name}.{table_name}").values[0][0]
        print(f"number of rows in {table_name}: {number_of_rows}")
        return columns


    def get_all_tables_info(self, sql_conn):
        """
        A method returning and printing all tables in database
        :parameter sql_conn instance of SQLConnection class
        :returns DF with all tables in database
        """
        val = sql_conn.read_to_df("SELECT TABLE_SCHEMA, TABLE_NAME FROM INFORMATION_SCHEMA.TABLES")
        print(self.format_string(str(val)))
        return val


    def preview_table(self, table_name, schema_name, nr_of_lines, sql_conn):
        """
        Preview method that prints first few rows of a table for users
        :parameter table_name name of table
        :parameter schema_name name of schema
        :parameter nr_of_lines number of lines to show
        :parameter sql_conn instance of SQLConnection class
        """
        df = sql_conn.read_to_df("SELECT * FROM " + schema_name + "." + table_name)
        print(self.format_string(table_name + "\n" + str(df.head(nr_of_lines))))
        return df.head(nr_of_lines)


    def push_data(self, table_name, schema_name, input_data, column_name_list, value_columns, crawldate_column, replace,
                  ignore_duplicates_check, sql_conn, recipients):
        """
        Method for pushing data to the desired table in database
        :param crawldate_column:
        :param table_name: Name of table
        :param schema_name: Name of schema
        :param input_data: Data to be pushed
        :param column_name_list: Used for giving correct name to data. Only required when input_data is a list
        :param value_columns: Columns that contains values. Used to check for identical lines and replaced based on replace
        :param replace: Boolean deciding if old values should be replaced if there is a identical row (except for value_columns)
                If True new value will be kept, if False, old value will be kept
        :param sql_conn: Instance of SQLConnection class
        :param ignore_duplicates_check True if you want no duplicate check to run
        :param recipients string with emails that should receive email in case of errors
        :raises Exception if input_data on unrecognizable structure.
        """
        engine, conn = sql_conn.get_connection()
        if input_data is None:
            self.send_email(sql_conn, recipients, "Input data cannot be None / empty")
            raise Exception("Input_data can not be None")
        col_list = self.get_table_columns(table_name, schema_name,
                                     sql_conn).values.tolist()  # using this in case they are not using column_name_list
        if max(len(col_list), len(column_name_list)) <= len(value_columns):
            self.send_email(sql_conn, recipients, "Every column cannot be a value column")
            raise Exception("every column cant be a value column!")

        schema_created = False
        try:
            engine.execute(sal.schema.CreateSchema(schema_name), checkfirst=True)
            print("Created new schema in DB: " + schema_name)
            schema_created = True  # solve differently??
        except:
            print("Schema " + schema_name + " exists in DB")
        if any(isinstance(i, list) for i in input_data):  # list in list
            df = pd.DataFrame(input_data, columns=column_name_list)
            self.df_to_db(table_name, schema_name, df, value_columns, replace, schema_created, sql_conn, ignore_duplicates_check,
                     crawldate_column, recipients)

        elif isinstance(input_data, list):  # list
            if len(column_name_list) == len(input_data):
                df = pd.DataFrame([input_data], columns=column_name_list)  # list = row
            else:
                df = pd.DataFrame(input_data, columns=column_name_list)  # list = column
            self.df_to_db(table_name, schema_name, df, value_columns, replace, schema_created, sql_conn, ignore_duplicates_check,
                     crawldate_column, recipients)

        elif isinstance(input_data, pd.DataFrame):  # pandas.Dataframe
            self.df_to_db(table_name, schema_name, input_data, value_columns, replace, schema_created, sql_conn,
                     ignore_duplicates_check, crawldate_column, recipients)
        elif isinstance(input_data, np.ndarray):  # np.ndarray
            df = pd.DataFrame(input_data)
            self.df_to_db(table_name, schema_name, df, value_columns, replace, schema_created, sql_conn, ignore_duplicates_check,
                     crawldate_column, recipients)
        elif isinstance(input_data, dict):  # dict
            df = pd.DataFrame(input_data.items(), columns=column_name_list)
            self.df_to_db(table_name, schema_name, df, value_columns, replace, schema_created, sql_conn,
                     ignore_duplicates_check, crawldate_column, recipients)
        else:
            body = f"Did not recognize data structure for insert into {schema_name}.{table_name} crawler. " \
                   f"Please change data structure"
            if schema_created:
                self.drop_schema(schema_name, sql_conn)
            self.send_email(sql_conn, recipients, body)
            raise Exception("DID NOT RECOGNISE INPUT: " + str(input_data))



    def df_to_db(self, table_name, schema_name, df, value_columns, replace, schema_created, sql_conn, ignore_duplicates_check,
                 crawldate_column, recipients):
        """Push Data to the selected table and provide errors if they occur,
        this should also append today's date to the data as a 'Crawled Date' column
        :param table_name: Name of table
        :param schema_name: Name of schema
        :param df: Dataframe to be pushed to database
        :param value_columns: columns with values
        :param replace: Boolean deciding if old or new values should be kept in case of duplicates (excluding value columns)
        :param schema_created: Schema_created flag for deleting schema if upload to DB fails. Used to avoid empty schemas in DB
        :param sql_conn: Intance of SQLConnection class
        :param ignore_duplicates_check True if you want no duplicate check to run
        :param crawldate_column: String describing name of crawl_date column. Default is 'CrawlDate'
        :param recipients string with emails that should receive email in case of errors
        """
        col_list = []  # used for feedback to user
        if crawldate_column:
            if crawldate_column not in df:
                df[crawldate_column] = pandas.Timestamp.now()
        else:  # not necessary with if else
            if "CrawlDate" not in df:
                df["CrawlDate"] = pandas.Timestamp.now()
        engine, conn = sql_conn.get_connection()
        varchar_cols = {}
        for i in range(len(df.columns)):
            if "object" in str(df.dtypes[i]):  # to avoid varchar(max)
                varchar_cols[df.columns[i]] = sal.types.NVARCHAR(self.max_varchar)
        before = self.get_row_count(table_name, schema_name, sql_conn)
        try:
            # this will fail if there is a new column
            df.to_sql(table_name, schema=schema_name, con=engine, if_exists="append", index=False, dtype=varchar_cols,
                      chunksize=int(math.floor(2100 / len(df.columns))) - 1, method='multi')
        except:
            try:
                cols = self.get_table_columns(table_name, schema_name, sql_conn)["COLUMN_NAME"].tolist()
                for i in range(len(df.columns.values)):
                    if df.columns.values[i] not in cols:
                        col_list.append(df.columns.values[i])
                        type = self.get_dtype(df.dtypes[i])
                        query = f"ALTER TABLE {schema_name}.{table_name} ADD [{df.columns.values[i]}] {type}"
                        sql_conn.execute(query)
                df.to_sql(table_name, schema=schema_name, con=engine, if_exists="append", index=False,
                          chunksize=int(math.floor(2100 / len(df.columns))) - 1, method='multi')
            except Exception as e:
                if schema_created:
                    self.drop_schema(schema_name, sql_conn)
                self.monitor_crawler(table_name, schema_name, 0, False, engine)
                self.send_email(sql_conn, recipients,
                          f"Adding to {schema_name}.{table_name} failed. The following error occured: " + str(e))
                raise Exception(f"Adding to {schema_name}.{table_name} failed. The following error occured: " + str(e))
        if not ignore_duplicates_check:
            self.remove_duplicates(table_name, schema_name, replace, value_columns, sql_conn, crawldate_column)
        after = self.get_row_count(table_name, schema_name, sql_conn)
        self.monitor_crawler(table_name, schema_name, after - before, True, engine)
        if len(col_list) != 0:
            print(
                f"{after - before} rows added to: {table_name}. Before: {before} rows, after: {after} rows. Also added following new columns: " + " ,".join(
                    col_list))
        else:
            print(f"{after - before} rows added to: {table_name}. Before: {before} rows, after: {after} rows.")


    def remove_duplicates(self, table_name, schema_name, replace, value_columns, sql_conn, crawldate_column='CrawlDate'):
        """
        Method for removing duplicates from DB
        :param table_name: Name of table
        :param schema_name: Name of schema
        :param replace: Boolean describing if old or new values are to be kept in case of duplicates.
        :param value_columns: List with name of columns where values are
        :param sql_conn: Instance of SQLConnection class
        :param crawldate_column: String describing name of crawl_date column. Default is 'CrawlDate'
        """
        if replace:
            sql_val = "Desc"
        else:
            sql_val = "Asc"
        col_list = self.get_table_columns(table_name, schema_name, sql_conn).values.tolist()
        if len(col_list) <= len(value_columns):
            raise Exception("every column cant be a value column!")
        col_string = ""
        server_name, db_name = sql_conn.get_names()
        for i in col_list:
            if i[0] in value_columns or i[0] == "CrawlDate":
                continue
            else:
                col_string += f"[{i[0]}], "
        params = f"BEGIN WITH CTE AS (SELECT *, ROW_NUMBER() OVER" \
                 f" (PARTITION BY {col_string[:-2]} order by [{crawldate_column}] {sql_val})" \
                 f" AS RN FROM [{db_name}].[{schema_name}].[{table_name}]) DELETE FROM CTE WHERE RN<>1 END COMMIT"
        sql_conn.execute(params)


    def get_dtype(self, pandas_type):
        """
        method used for finding correct type
        :param pandas_type: Type in pandas
        :return: Type in SQL
        """
        if "object" in str(pandas_type):
            return sal.types.NVARCHAR(self.max_varchar)
        elif "datetime" in str(pandas_type):
            return sal.types.DateTime()
        elif "float" in str(pandas_type):
            return sal.types.Float(asdecimal=True)
        elif "int" in str(pandas_type):
            return sal.types.INT()
        else:  # rarely happen?
            return sal.types.NVARCHAR(self.max_varchar)


    def format_string(self, string):
        """
        for formatting strings
        :param string: string to be formatted

        :return: formatted string
        """
        s = "-------------------------------------------------------"
        return s + "\n" + string + "\n" + s + "\n"


    def drop_schema(self, schema_name, sql_conn):
        """
        Method for dropping schema and all subtables
        :param schema_name: Name of schema
        :param sql_conn: Instance of SQLConnection class
        """
        df = sql_conn.read_to_df(f"SELECT TABLE_NAME FROM INFORMATION_SCHEMA.TABLES Where TABLE_SCHEMA = '{schema_name}'")
        for i in range(len(df)):
            sql_conn.execute(f"DROP TABLE [{schema_name}].[{df.values[i][0]}]")
        sql_conn.execute(f"DROP SCHEMA [{schema_name}]")
        print(f"{schema_name} and subtables has been dropped")


    def monitor_crawler(self, table_name, schema_name, nr_of_lines, successful, engine):
        """
        Method for running the monitoring table
        :param table_name: Name of table
        :param schema_name: Name of schema
        :param nr_of_lines number of lines added
        :param successful if upload was successful or not
        :param engine engine used to connect to database
        """
        try:
            engine.execute(sal.schema.CreateSchema("Info"), checkfirst=True)
            print("Info-Schema created")
        except:
            print("updating monitoring table...")
        col_dtypes = {
            "Table": sal.types.NVARCHAR(255),
            "Schema": sal.types.NVARCHAR(255),
            "Nr_of_rows": sal.types.INT(),
            "Timestamp": sal.types.DateTime()
        }
        df = pd.DataFrame([[schema_name, table_name, nr_of_lines, successful, pd.pandas.Timestamp.now()]],
                          columns=["Schema", "Table", "Nr_of_rows", "Upload successful", "Timestamp"])
        df.to_sql("MonitoringTable", schema="Info", con=engine, if_exists="append", index=False, dtype=col_dtypes)


    def get_row_count(self, table_name, schema_name, sql_conn):
        """
        Method for getting row count
        :param table_name: Name of table
        :param schema_name: Name of schema
        :param sql_conn: Instance of SQLConnection class
        :return number of lines in table
        """
        # this might later be replaced by count=df.to_sql(). Not a current feature, but talks of implementing it.
        try:
            df = sql_conn.read_to_df(f"SELECT COUNT(*) FROM [{schema_name}].[{table_name}]")
            return int(df.values[0][0])
        except:
            return 0


    def send_email(self, sql_conn, recipients, body,subject):
        """
        :param sql_conn: Instance of SQLConnection class
        :param recipients recipients of email, string or list
        :param body message to be written in email. The subject is always "Crawler failed"
        """
        if recipients is not None:
            engine, conn = sql_conn.get_connection()
            if isinstance(recipients, list):
                recipients = ";".join(recipients)
            query = "Begin EXEC msdb.dbo.sp_send_dbmail " \
                    "@profile_name = 'TechnologySupport', " \
                    f"@recipients = '{recipients}', " \
                    f"@body = '{body}', " \
                    f"@subject = '{subject}'; End Commit"
            engine.execute(query)
            print(f"mail sent to {recipients}")

    def set_max_varchar(self,varchar_size):
        """
        :param max_varchar_size size of maximum length of varchar
        """
        if varchar_size > 1023:
            raise Exception("Varchar too big, maximum is 1023")
        elif varchar_size == 0:
            raise Exception("Varchar size cannot be 0")
        else:
            self.max_varchar = varchar_size

    def get_newest_date(self, table_name, schema_name, col_name, sql_conn):
        """
        Method for getting the newest date present 
        :param table_name: Name of table
        :param schema_name: Name of schema
        :param sql_conn: Instance of SQLConnection class
        :return number of lines in table
        """
        try:

            res_df = sql_conn.read_to_df(f"Select Top 1 {col_name} From {schema_name}.{table_name} Order By {col_name} Desc")
            date = pd.to_datetime(res_df.iloc[0][0])
            return date
        except:
            return 0
